@@ -15,6 +15,8 @@
 #include "atr.h"
 #include <stdlib.h>
 #include <string.h>
+#include <sys/stat.h>
+#include <unistd.h>
 
 #define CCID_ERR_ABORTED	0xFF	/* CMD ABORTED */
 #define CCID_ERR_ICC_MUTE	0xFE
@@ -217,6 +219,9 @@ typedef struct ccid_status {
 	unsigned char *sbuf[OPENCT_MAX_SLOTS];
 	size_t slen[OPENCT_MAX_SLOTS];
 	unsigned char seq;
+	int support_events;
+	int events_active;
+	ifd_usb_capture_t *event_cap;
 } ccid_status_t;
 
 static int ccid_checkresponse(void *status, int r)
@@ -332,8 +337,10 @@ static int ccid_command(ifd_reader_t * reader, const unsigned char *cmd,
 		ifd_debug(3, "sending:%s", ct_hexdump(cmd, cmd_len));
 
 	rc = ifd_device_send(reader->device, cmd, cmd_len);
-	if (rc < 0)
+	if (rc < 0) {
+		ifd_debug(1, "ifd_device_send failed %d", rc);
 		return rc;
+	}
 	while (1) {
 		rc = ifd_device_recv(reader->device, res, req_len, 10000);
 		if (rc < 0)
@@ -474,7 +481,7 @@ static int ccid_open_usb(ifd_device_t * dev, ifd_reader_t * reader)
 {
 	ccid_status_t *st;
 	ifd_device_params_t params;
-	int r, i, c, ifc, alt;
+	int r, i, c, ifc, alt, num_alt;
 	struct ifd_usb_device_descriptor de;
 	struct ifd_usb_config_descriptor conf;
 	struct ifd_usb_interface_descriptor *intf;
@@ -483,6 +490,14 @@ static int ccid_open_usb(ifd_device_t * dev, ifd_reader_t * reader)
 	struct force_parse_device_st *force_dev;
 	unsigned char *_class;
 	unsigned char *p;
+	int support_events = 0;
+
+#define PCSCLITE_FILE "/var/run/pcscd/pcscd.comm"
+	struct stat buf;
+
+	/* give priority to pcsc-lite for CCID devices */
+	if (0 == stat(PCSCLITE_FILE, &buf))
+		sleep(3);
 
 	if (ifd_usb_get_device(dev, &de)) {
 		ct_error("ccid: device descriptor not found");
@@ -512,8 +527,8 @@ static int ccid_open_usb(ifd_device_t * dev, ifd_reader_t * reader)
 			continue;
 
 		for (ifc = 0; ifc < conf.bNumInterfaces; ifc++) {
-			for (alt = 0; alt < conf.interface[ifc].num_altsetting;
-			     alt++) {
+			num_alt = conf.interface[ifc].num_altsetting;
+			for (alt = 0; alt < num_alt; alt++) {
 				int typeok = 0;
 				int ok = 0;
 				intf = &conf.interface[ifc].altsetting[alt];
@@ -536,6 +551,9 @@ static int ccid_open_usb(ifd_device_t * dev, ifd_reader_t * reader)
 					params.usb.ep_intr = 0;
 					ok |= 4;
 				}
+				if (intf->bNumEndpoints == 3) {
+					support_events = 1;
+				}
 				for (i = 0; i < intf->bNumEndpoints; i++) {
 					if (((intf->endpoint[i].bmAttributes &
 					      IFD_USB_ENDPOINT_TYPE_MASK) ==
@@ -581,8 +599,36 @@ static int ccid_open_usb(ifd_device_t * dev, ifd_reader_t * reader)
 			if (!intf)
 				continue;
 			if (!intf->extralen) {
-				intf = NULL;
-				continue;
+				int i;
+				/* Buggy O2 Micro CCID SC Reader has zero extra len at interface level but not endpoint descriptor.
+				 * Patch the interface level field and proceed.
+				 * ProdID 7762 reader is in Dell Latitude D620 and 7772 is in D630.
+				 */
+        			if( de.idVendor == 0x0b97 && (de.idProduct == 0x7762 || de.idProduct == 0x7772) )  {
+					ct_error("ccid: extra len is zero, patching O2 Micro support");
+					for (i=0; i<intf->bNumEndpoints; i++)  {
+						/* find the extra[] array */
+						if( intf->endpoint[i].extralen == 54 )  {
+							/* get the extra[] from the endpoint */
+							intf->extralen = 54;
+							/* avoid double free on close, allocate here */
+							intf->extra = malloc(54);
+							if( intf->extra )  {
+								memcpy( intf->extra, intf->endpoint[i].extra, 54 );
+								break;
+							}
+							else  {
+								intf = NULL;
+								continue;
+							}
+						}
+					}
+				}
+				else  {
+					intf = NULL;
+					ct_error("ccid: extra len is zero, continuing");
+					continue;
+				}
 			}
 
 			r = intf->extralen;
@@ -639,7 +685,10 @@ static int ccid_open_usb(ifd_device_t * dev, ifd_reader_t * reader)
 		params.usb.configuration = -1;
 
 	params.usb.interface = intf->bInterfaceNumber;
-	params.usb.altsetting = intf->bAlternateSetting;
+	if (num_alt > 1 || intf->bAlternateSetting > 0)
+		params.usb.altsetting = intf->bAlternateSetting;
+	else
+		params.usb.altsetting = -1;
 
 	r = ccid_parse_descriptor(&ccid, p, r - i);
 	ifd_usb_free_configuration(&conf);
@@ -753,7 +802,10 @@ static int ccid_open_usb(ifd_device_t * dev, ifd_reader_t * reader)
 		reader->nslots++;	/* one virtual slot for RFID escape */
 		st->proto_support |= SUPPORT_ESCAPE;
 	}
-	ifd_debug(3, "Accepted %04x:%04x with features 0x%x and protocols 0x%x", de.idVendor, de.idProduct, ccid.dwFeatures, ccid.dwProtocols);
+
+	st->support_events = support_events;
+
+	ifd_debug(3, "Accepted %04x:%04x with features 0x%x and protocols 0x%x events=%d", de.idVendor, de.idProduct, ccid.dwFeatures, ccid.dwProtocols, st->support_events);
 	return 0;
 }
 
@@ -805,6 +857,23 @@ static int ccid_open(ifd_reader_t * reader, const char *device_name)
 	}
 }
 
+/*
+ * Close the device
+ */
+static int ccid_close(ifd_reader_t * reader)
+{
+	ccid_status_t *st = (ccid_status_t *) reader->driver_data;
+	
+	ifd_debug(1, "called.");
+
+	if (st->event_cap != NULL) {
+		ifd_usb_end_capture(reader->device, st->event_cap);
+		st->event_cap = NULL;
+	}
+
+	return 0;
+}
+
 static int ccid_activate(ifd_reader_t * reader)
 {
 	ifd_debug(1, "called.");
@@ -881,20 +950,29 @@ static int ccid_card_status(ifd_reader_t * reader, int slot, int *status)
 			return 0;
 		}
 	}
+
 	r = ccid_prepare_cmd(reader, cmdbuf, 10, 0, CCID_CMD_GETSLOTSTAT,
 			     NULL, NULL, 0);
 	if (r < 0)
 		return r;
 	r = ccid_command(reader, cmdbuf, 10, ret, 10);
-	if (r < 0)
-		return r;
-	switch (ret[7] & 3) {
-	case 2:
+	if (r == IFD_ERROR_NO_CARD) {
 		stat = 0;
-		break;
-	default:
-		stat = IFD_CARD_PRESENT;
 	}
+	else if (r < 0) {
+		return r;
+	}
+	else {
+		switch (ret[7] & 3) {
+		case 2:
+			stat = 0;
+			break;
+		default:
+			stat = IFD_CARD_PRESENT;
+			break;
+		}
+	}
+
 	ifd_debug(1, "probed result: %d", IFD_CARD_STATUS_CHANGED | stat);
 
 	*status = IFD_CARD_STATUS_CHANGED | stat;
@@ -1110,7 +1188,7 @@ static int ccid_set_protocol(ifd_reader_t * reader, int s, int proto)
 	}
 
 	if ((st->flags & FLAG_NO_SETPARAM) == 0 &&
-		((st->flags & FLAG_AUTO_ATRPARSE) == 0 |
+		((st->flags & FLAG_AUTO_ATRPARSE) == 0 ||
 		proto != IFD_PROTOCOL_T0)) {
 
 		/* if FLAG_AUTO_ATRPARSE, only set the protocol. */
@@ -1282,6 +1360,117 @@ static int ccid_recv(ifd_reader_t * reader, unsigned int dad,
 	return r;
 }
 
+static int ccid_before_command(ifd_reader_t * reader)
+{
+	ccid_status_t *st = (ccid_status_t *) reader->driver_data;
+	int rc;
+
+	ifd_debug(1, "called.");
+
+	if (!st->events_active) {
+		return 0;
+	}
+
+	if (st->event_cap == NULL) {
+		return 0;
+	}
+
+	rc = ifd_usb_end_capture(reader->device, st->event_cap);
+	st->event_cap = NULL;
+
+	return rc;
+}
+
+static int ccid_after_command(ifd_reader_t * reader)
+{
+	ccid_status_t *st = (ccid_status_t *) reader->driver_data;
+
+	ifd_debug(1, "called.");
+
+	if (!st->events_active) {
+		return 0;
+	}
+
+	if (st->event_cap != NULL) {
+		return 0;
+	}
+
+	return ifd_usb_begin_capture(
+		reader->device,
+		IFD_USB_URB_TYPE_INTERRUPT,
+		reader->device->settings.usb.ep_intr,
+		8, &st->event_cap
+	);
+}
+
+static int ccid_get_eventfd(ifd_reader_t * reader, short *events)
+{
+	ccid_status_t *st = (ccid_status_t *) reader->driver_data;
+	int fd;
+
+	ifd_debug(1, "called.");
+
+	if (!st->support_events) {
+		return -1;
+	}
+
+	fd = ifd_device_get_eventfd(reader->device, events);
+
+	if (fd != -1) {
+		st->events_active = 1;
+	}
+
+	return fd;
+}
+
+static int ccid_event(ifd_reader_t * reader, int *status, size_t status_size)
+{
+	ccid_status_t *st = (ccid_status_t *) reader->driver_data;
+	unsigned char ret[20];
+	int bytes;
+
+	ifd_debug(1, "called.");
+
+	if (status_size < reader->nslots) {
+		return IFD_ERROR_BUFFER_TOO_SMALL;
+	}
+
+	bytes = ifd_usb_capture_event(reader->device, st->event_cap, ret, 8);
+	if (bytes < 0) {
+		return bytes;
+	}
+
+	if (bytes > 0 && ret[0] == 0x50) {
+		int slot;
+		ifd_debug(3, "status received:%s", ct_hexdump(ret, bytes));
+		for (slot=0;slot<reader->nslots;slot++) {
+			if (1 + (slot / 4) < bytes) {
+				int bits = (ret[1 + (slot / 4)] >> (2 * (slot % 4))) & 0x3;
+				if (bits & 2)
+					status[slot] |= IFD_CARD_STATUS_CHANGED;
+				if (bits & 1)
+					status[slot] |= IFD_CARD_PRESENT;
+				else
+					status[slot] &= ~IFD_CARD_PRESENT;
+
+				ifd_debug(1, "slot %d event result: %08x", slot, status[slot]);
+				st->icc_present[slot] = status[slot] & IFD_CARD_PRESENT;
+			}
+		}
+	}
+
+	return 0;
+}
+
+static int ccid_error(ifd_reader_t * reader)
+{
+	(void)reader;
+
+	ifd_debug(1, "called.");
+
+	return IFD_ERROR_DEVICE_DISCONNECTED;
+}
+
 /*
  * Driver operations
  */
@@ -1293,6 +1482,7 @@ static struct ifd_driver_ops ccid_driver;
 void ifd_ccid_register(void)
 {
 	ccid_driver.open = ccid_open;
+	ccid_driver.close = ccid_close;
 	ccid_driver.activate = ccid_activate;
 	ccid_driver.deactivate = ccid_deactivate;
 	ccid_driver.card_status = ccid_card_status;
@@ -1302,6 +1492,11 @@ void ifd_ccid_register(void)
 	ccid_driver.send = ccid_send;
 	ccid_driver.recv = ccid_recv;
 	ccid_driver.escape = ccid_escape;
+	ccid_driver.before_command = ccid_before_command;
+	ccid_driver.after_command = ccid_after_command;
+	ccid_driver.get_eventfd = ccid_get_eventfd;
+	ccid_driver.event = ccid_event;
+	ccid_driver.error = ccid_error;
 
 	ifd_driver_register("ccid", &ccid_driver);
 }
